import AlteryxPythonSDK as Sdk
import connectors_core.alteryx.helpers as alteryx_helpers

import google.cloud.bigquery

import re

import datetime


class SchemaConversionException(Exception):
    def __init__(self, message):
        self.message = message


class BigQuerySchema:
    def __init__(self, table: google.cloud.bigquery.table.Table, max_repeated_records: int, max_field_length: int):
        self.table = table
        self.max_repeated_records = max_repeated_records
        self.max_field_length = max_field_length
        self.has_repeated_records = False
        for field in self.table.schema:
            if field.mode == 'REPEATED' and field.field_type == 'RECORD':
                self.has_repeated_records = True

    def get_record_info(self) -> Sdk.RecordInfo:
        """
        Wraps get_sdk_schema with alteryx_helpers.build_record_info
        :return: An instance of Sdk.RecordInfo for flattened representation of BigQuery table schema
        """
        return alteryx_helpers.build_record_info(self.alteryx_engine, self.get_sdk_schema())

    def get_sdk_schema(self, as_re=False):
        """
        Builds a list of dictionaries specifying the Alteryx SDK flattened representation of the BigQuery table schema
        :return: A list of dictionaries containing (description, name, Sdk.FieldType, size, scale) keys for each flattened field
        """
        sdk_schema = []
        for field in self.table.schema:
            if field.field_type == 'RECORD':
                for subfield in field.fields:
                    if field.mode == 'REPEATED':
                        if not as_re:
                            for idx in range(self.max_repeated_records):
                                sdk_schema.append({
                                    'description': subfield.description,
                                    'name': self._format_field_name(field, subfield=subfield, id=idx),
                                    **self._get_sdk_type(subfield.field_type)
                                })
                        else:
                            sdk_schema.append({
                                'description': subfield.description,
                                'name': self._format_field_name(field, subfield=subfield, id='\d+'),
                                **self._get_sdk_type(subfield.field_type)
                            })
                    else:
                        sdk_schema.append({
                            'description': subfield.description,
                            'name': self._format_field_name(field, subfield=subfield),
                            **self._get_sdk_type(subfield.field_type)
                        })
            else:
                sdk_schema.append({
                    'description': field.description,
                    'name': self._format_field_name(field),
                    **self._get_sdk_type(field.field_type)
                })
        return sdk_schema

    def expected_field_names_for_schema(self, as_re=False):
        if not (self.table and self.table.schema):
            return []
        return [field['name'] for field in self.get_sdk_schema(as_re)]

    def _format_field_name(self, field, subfield=None, id=None):
        """
        Helper for translating BigQuery column names into flattened representation column names
        :param field: the base field
        :param subfield: the nested field, if exists
        :param id: the repetition id for a repeated record, if exists
        :return: a string column name for flattened representation of nested, repeated records
        """
        if subfield is None and id is None:
            return field.name
        elif id is None:
            return 'nr__{}__{}'.format(field.name, subfield.name)
        return 'nrr__{}__{}__{}'.format(field.name, id, subfield.name)


    def _get_sdk_type(self, cloud_type):
        """
        Helper for converting Big Query types to Alteryx SDK Types
        :param cloud_type: str representation of Big Query Google Cloud Type
        :return: dictionary representation of Alteryx SDK Type
        """
        try:
            return {
                'STRING': {'type': Sdk.FieldType.v_wstring, 'size': self.max_field_length, 'scale': 0},
                'BYTES': {'type': Sdk.FieldType.blob, 'size': self.max_field_length, 'scale': 0},
                'INTEGER': {'type': Sdk.FieldType.int64, 'size': 0, 'scale': 0},
                'FLOAT': {'type': Sdk.FieldType.float, 'size': 0, 'scale': 0},
                'BOOLEAN': {'type': Sdk.FieldType.bool, 'size': 0, 'scale': 0},
                'TIMESTAMP': {'type': Sdk.FieldType.datetime, 'size': 0, 'scale': 0},
                'TIME': {'type': Sdk.FieldType.time, 'size': 0, 'scale': 0},
                'DATE': {'type': Sdk.FieldType.date, 'size': 0, 'scale': 0},
                'DATETIME': {'type': Sdk.FieldType.datetime, 'size': 0, 'scale': 0},
                'RECORD': {'type': Sdk.FieldType.blob, 'size': self.max_field_length, 'scale': 0},
            }[cloud_type]
        except KeyError as e:
            return {'type': Sdk.FieldType.wstring, 'size': self.max_field_length, 'scale': 0}


class BigQuerySchemaWriter(BigQuerySchema):
    def __init__(self, table, max_repeated_records, max_field_length, record_info: Sdk.RecordInfo):
        super().__init__(table, max_repeated_records, max_field_length)
        self.record_info = record_info

    def _infer_field_options_from_name(self, field_name):
        """
        Helper for inferring if a field is nested or nested, repeated from the string field name
        :param field_name: A string representation of the flattened column names
        :return: A dictionary of booleans with keys (isNested, isRepeated)
        """
        nrr_re = re.compile('nrr__\w+__\d+__\w+')
        if nrr_re.fullmatch(field_name):
            return {'isNested': True, 'isRepeated': True}
        nr_re = re.compile('nr__\w+__\w+')
        if nr_re.fullmatch(field_name):
            return {'isNested': True, 'isRepeated': False}
        return {'isNested': False, 'isRepeated': False}

    def is_expected(self, name, regexes):
        combined = '(?:%s)' % '|'.join(regexes)
        return re.match(combined, name) is not None

    def build_row_from_record(self, record_ref, alteryx_engine: Sdk.AlteryxEngine = None):
        my_data = {}
        try:
            for i in range(0, self.record_info.num_fields):
                field = self.record_info[i]  # type: Sdk.Field
                if not self.is_expected(field.name, self.expected_field_names_for_schema(as_re=True)):
                    if alteryx_engine:
                        raise SchemaConversionException(alteryx_engine.xmsg(
                            'Unexpected field name for destination table schema: @1', field.name
                        ))
                    else:
                        raise SchemaConversionException(
                            'Unexpected field name for destination table schema: {}'.format(field.name)
                        )
                column_options = self._infer_field_options_from_name(field.name)
                field_val = alteryx_helpers.get_value_for_field(field, record_ref)
                if type(field_val) is bytearray:
                    field_val = field_val.decode("utf-8")
                if type(field_val) is datetime.timedelta:
                    field_val = str(field_val)
                if field.type == Sdk.FieldType.date:
                    field_val = str(field_val).split(' ')[0]
                if column_options['isNested']:
                    if column_options['isRepeated']:
                        _, top_attr, index, nested_attr = field.name.split('__')
                        index = int(index)
                        if top_attr not in my_data.keys():
                            my_data[top_attr] = []
                        try:
                            my_data[top_attr][index][nested_attr] = field_val
                        except IndexError:
                            my_data[top_attr].insert(index, {nested_attr: field_val})
                    else:
                        _, top_attr, nested_attr = field.name.split('__')
                        if top_attr not in my_data.keys():
                            my_data[top_attr] = { nested_attr: field_val }
                        else:
                            my_data[top_attr][nested_attr] = field_val
                else:
                    my_data[field.name] = field_val
            return self._remove_empty_repetitions(my_data)
        except KeyError as e:
            if alteryx_engine:
                raise SchemaConversionException(alteryx_engine.xmsg(
                    'Schema Error encountered while building a BigQuery Row object from Designer Record.'
                ))
            else:
                raise SchemaConversionException('Schema Error encountered while building a BigQuery Row object from Designer Record.')
        except ValueError as e:
            if alteryx_engine:
                raise SchemaConversionException(alteryx_engine.xmsg(
                    'Schema Error encountered while building a BigQuery Row object from Designer Record.'
                ))
            else:
                raise SchemaConversionException('Schema Error encountered while building a BigQuery Row object from Designer Record.')
        except IndexError as e:
            if alteryx_engine:
                raise SchemaConversionException(alteryx_engine.xmsg(
                    'Schema Error encountered while building a BigQuery Row object from Designer Record. Did not receive enough fields.'
                ))
            else:
                raise SchemaConversionException(
                    'Schema Error encountered while building a BigQuery Row object from Designer Record. Did not receive enough fields.'
                )

    def _remove_empty_repetitions(self, payload):
        """
        Remove any nested, repeated records that are all null
        :param payload: a dictionary representation of a row in the BigQuery table
        :return: the dictionary stripped of any null repetitions
        """
        for field in payload:
            indices_to_remove = []
            if type(payload[field]) is list:
                for i, nested_record in enumerate(payload[field]):
                    empty_record = True
                    for nested_field in nested_record:
                        empty_record = empty_record and nested_record[nested_field] is None
                    if empty_record:
                        indices_to_remove.append(i)
            for index in indices_to_remove[::-1]:
                payload[field].pop(index)
        return payload


class BigQuerySchemaReader(BigQuerySchema):
    def __init__(self,
                 alteryx_engine: Sdk.AlteryxEngine,
                 table: google.cloud.bigquery.table.Table,
                 max_repeated_records: int,
                 max_field_length: int):
        super().__init__(table, max_repeated_records, max_field_length)
        self.alteryx_engine = alteryx_engine

    def generate_records_from_field(self, result, field, record_info_out):
        """
        Yields as many (record_field, field_value) pairs as necessary for the given field

        If field is not nested or repeated, then yield one pair.
        If the field is nested and not repeated, then yield as many pairs as there are nested attributes on that field
        If the field is nested and repeated, then yield (self.max_repeated_records * number of nested attributes) pairs
        :param result: An instance of google.cloud.bigquery.table.Row; provides field values
        :param field: An instance of google.cloud.bigquery.schema.SchemaField; the field to generate a record for
        :param record_info_out: An instance of Sdk.RecordInfo
        :return: (record_field, field_value) pairs
        """
        if field.field_type == 'RECORD':
            for subfield in field.fields:
                if field.mode == 'REPEATED':
                    for i in range(self.max_repeated_records):
                        idx = record_info_out.get_field_num(self._format_field_name(field, subfield, i))
                        record_field = record_info_out[idx]
                        try:
                            nested_field = getattr(result, field.name)[i]
                            field_value = nested_field[subfield.name]
                        except IndexError as e:
                            # nested_field does not have enough repetitions
                            field_value = None
                        yield record_field, field_value
                else:
                    idx = record_info_out.get_field_num(self._format_field_name(field, subfield))
                    record_field = record_info_out[idx]
                    field_value = getattr(result, field.name)[subfield.name]
                    yield record_field, field_value
        else:
            idx = record_info_out.get_field_num(self._format_field_name(field))
            record_field = record_info_out[idx]
            field_value = getattr(result, field.name)
            yield record_field, field_value

