import logging

from oauthlib.oauth2 import BackendApplicationClient
import AlteryxPythonSDK as Sdk

from connectors_core.oauth2 import OAUTH_CODE_WORKFLOW, OAUTH_CLIENT_CREDENTIALS_WORKFLOW
from odata import ODataService
from odata.entity import EntityBase
from odata.query import Query
from requests_oauthlib import OAuth2Session
from connectors_core.oauth2.clients import AzureActiveDirectoryWebApplicationClient, \
    AzureActiveDirectoryBackendApplicationClient

from typing import Type

DATA_SERVICE_PATH = 'api/data/v9.0/'
AUTHORIZATION_SERVER_URL = 'https://login.microsoftonline.com'


class DynamicsCrmApiClient:
    logger = logging.getLogger('crm.DynamicsCrmApiClient')

    def __init__(
            self,
            org_uri: str,
            tenant: str,
            client_id: str,
            client_secret: str,
            auth_token: dict,
            metadata: str,
            auth_mode: str = OAUTH_CODE_WORKFLOW,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        if not org_uri[-1] == '/':
            org_uri = org_uri + '/'
        if auth_mode == OAUTH_CODE_WORKFLOW:
            refresh_kwargs = {}
            client = AzureActiveDirectoryWebApplicationClient(
                client_id=client_id,
                client_secret=client_secret,
                token=auth_token
            )
        elif auth_mode == OAUTH_CLIENT_CREDENTIALS_WORKFLOW:
            refresh_kwargs = {'resource': org_uri}
            client = AzureActiveDirectoryBackendApplicationClient(
                client_id=client_id,
                client_secret=client_secret,
                alteryx_engine=alteryx_engine
            )
        else:
            if not alteryx_engine:
                raise AttributeError("Unsupported auth_mode: {}".format(auth_mode))
            else:
                raise AttributeError(alteryx_engine.xmsg("Unsupported auth_mode: @1", auth_mode))

        session = OAuth2Session(
            client_id=client_id,
            client=client,
            token=auth_token,
            auto_refresh_url="{}/{}/oauth2/token".format(AUTHORIZATION_SERVER_URL, tenant),
            auto_refresh_kwargs=refresh_kwargs,
            token_updater=self._update_token
        )
        self.service = ODataService("{}{}".format(org_uri, DATA_SERVICE_PATH), session=session, reflect_entities=True,
                                    metadata=metadata)

    def get_entity_cls(self, entity_name: str) -> Type[EntityBase]:
        return self.service.entities[entity_name] if entity_name in self.service.entities else None

    def query(self, entity: Type[EntityBase]) -> Query:
        return self.service.query(entity)

    def save(self, entity: EntityBase):
        self.service.save(entity, force_refresh=False)

    def delete(self, entity: EntityBase):
        self.service.delete(entity)

    def _update_token(self, token: dict):
        self.logger.info("Token updated: expires_at={}".format(token['expires_at']))
