import AlteryxPythonSDK as Sdk
import json
from typing import Union, Any, Callable, IO

import fastavro

from .errors import DataTypeNotSupportedError
import connectors_core.alteryx.helpers as alteryx_helpers
from .interfaces import FileInputReader


def unsupported_type(field_type: Sdk.FieldType, alteryx_engine: Sdk.AlteryxEngine = None):
    def raise_unsupported_error(record_ref: Sdk.RecordRef=None):
        if alteryx_engine:
            raise DataTypeNotSupportedError(alteryx_engine.xmsg(
                "Data Type '@1' is not supported by the PythonSDK", field_type
            ))
        else:
            raise DataTypeNotSupportedError(
                "Data Type '{}' is not supported by the PythonSDK".format(field_type))

    return raise_unsupported_error


def get_sdk_type_for_avro_type(
        avro_type: Union[str, dict, list],
        max_field_size: int = 255,
        alteryx_engine: Sdk.AlteryxEngine = None) -> (Sdk.FieldType, int, int):
    """
    see https://avro.apache.org/docs/1.8.2/spec.html#schemas

    see https://help.alteryx.com/11.7/index.htm#Avro_Data_Types.htm

    :param max_field_size: Max field Size value
    :param avro_type: The name of the Avro type
    :return:  3-Tuple containing the Sdk Type, the size, and the scale
    """

    if avro_type == 'null':
        if alteryx_engine:
            raise DataTypeNotSupportedError(alteryx_engine.xmsg('Null field type is not supported'))
        else:
            raise DataTypeNotSupportedError('Null field type is not supported')
    elif isinstance(avro_type, list):
        # Union type
        if len(avro_type) > 2:
            # Invalid unions are imported as JSON into an V_WString
            return Sdk.FieldType.v_wstring, max_field_size, 0
        if 'null' in avro_type:
            avro_type.remove('null')
            avro_type = avro_type[0]
    elif isinstance(avro_type, dict):
        field_type = avro_type['type']
        if field_type == 'enum':
            return Sdk.FieldType.v_wstring, max_field_size, 0
        elif field_type in ['record', 'array', 'map']:
            # The following Avro types are not supported natively, but are imported as JSON into a String
            return Sdk.FieldType.v_wstring, max_field_size, 0

    try:
        return {
            'string': (Sdk.FieldType.v_wstring, max_field_size, 0),
            'long': (Sdk.FieldType.int64, 0, 0),
            'int': (Sdk.FieldType.int32, 0, 0),
            'bytes': (Sdk.FieldType.blob, 0, 0),
            'float': (Sdk.FieldType.float, 0, 0),
            'double': (Sdk.FieldType.double, 0, 0),
            'boolean': (Sdk.FieldType.bool, 0, 0),
            'fixed': (Sdk.FieldType.blob, 0, 0),
            'enum': (Sdk.FieldType.v_wstring, max_field_size)
        }[avro_type]
    except KeyError:
        return Sdk.FieldType.v_wstring, max_field_size, 0


def get_avro_type_for_sdk_type(sdk_type: Sdk.FieldType,
                               size: int = 0,
                               scale: int = 0,
                               support_nulls: bool = False,
                               alteryx_engine: Sdk.AlteryxEngine = None) -> [str, dict]:
    """
    see https://avro.apache.org/docs/1.8.2/spec.html#schemas

    :return:  type name or dict for complex types
    """
    try:
        avro_type = {
            Sdk.FieldType.v_wstring: 'string',
            Sdk.FieldType.int64: 'long',
            Sdk.FieldType.int32: 'int',
            Sdk.FieldType.int16: 'int',
            Sdk.FieldType.blob: 'bytes',
            Sdk.FieldType.float: 'float',
            Sdk.FieldType.double: 'double',
            Sdk.FieldType.bool: 'boolean',
            Sdk.FieldType.wstring: 'string',
            Sdk.FieldType.string: 'string',
            Sdk.FieldType.byte: 'int',
            Sdk.FieldType.date: 'string',
            Sdk.FieldType.datetime: 'string',
            Sdk.FieldType.spatialobj: 'bytes',
            Sdk.FieldType.fixeddecimal: 'double',
            Sdk.FieldType.time: 'string',
            Sdk.FieldType.v_string: 'string'
        }[sdk_type]
    except KeyError:
        if alteryx_engine:
            raise DataTypeNotSupportedError(alteryx_engine.xmsg(
                "Data Type @1 is not supported for conversion to Avro", sdk_type
            ))
        else:
            raise DataTypeNotSupportedError("Data Type {} is not supported for conversion to Avro".format(sdk_type))

    if support_nulls:
        # Unions ... are represented using JSON arrays. For example, ["null", "string"]
        avro_type = ['null', avro_type]

    return avro_type


def get_avro_value_from_field(
        field: Sdk.Field,
        record_ref: Sdk.RecordRef,
        data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
        alteryx_engine: Sdk.AlteryxEngine = None) -> Any:
    accessor = {
        Sdk.FieldType.string: field.get_as_string,
        Sdk.FieldType.wstring: field.get_as_string,
        Sdk.FieldType.v_wstring: field.get_as_string,
        Sdk.FieldType.v_string: field.get_as_string,
        Sdk.FieldType.int32: field.get_as_int32,
        Sdk.FieldType.double: field.get_as_double,
        Sdk.FieldType.float: field.get_as_double,
        Sdk.FieldType.bool: field.get_as_bool,
        Sdk.FieldType.datetime: field.get_as_string,
        Sdk.FieldType.date: field.get_as_string,
        Sdk.FieldType.fixeddecimal: field.get_as_double,
        Sdk.FieldType.time: field.get_as_string,
        Sdk.FieldType.blob: unsupported_type(Sdk.FieldType.blob, alteryx_engine),
        Sdk.FieldType.spatialobj: unsupported_type(Sdk.FieldType.spatialobj, alteryx_engine),
        Sdk.FieldType.byte: field.get_as_int32,
        Sdk.FieldType.int16: field.get_as_int32,
        Sdk.FieldType.int64: field.get_as_int64
    }[field.type]

    try:
        return accessor(record_ref)
    except TypeError as e:
        if data_conversion_error_handler:
            return data_conversion_error_handler(field, record_ref, e)
        else:
            raise


class AvroInputReader(FileInputReader):

    def __init__(
            self,
            stream: IO,
            output_anchor: Sdk.OutputAnchor,
            alteryx_engine: Sdk.AlteryxEngine,
            stream_length_bytes: Union[int, None] = None,
            field_source_resolver: Callable[[Sdk.Field], str] = lambda f: ''):
        super().__init__(stream, output_anchor, alteryx_engine, stream_length_bytes, field_source_resolver)

        self.reader = fastavro.reader(stream)

    def init(self):
        fields = []
        for field in self.reader.schema['fields']:
            sdk_type, size, scale = get_sdk_type_for_avro_type(field['type'], alteryx_engine=self.alteryx_engine)
            fields.append({
                'name': field['name'],
                'type': field['type'],
                'sdk_type': sdk_type,
                'size': size,
                'scale': scale,
                'description': field['doc'] if 'doc' in field else ''
            })
        self.record_info = alteryx_helpers.build_record_info(
            self.alteryx_engine,
            [{
                'name': f['name'],
                'type': f['sdk_type'],
                'size': f['size'],
                'scale': f['scale'],
                'description': f['description']
            } for f in fields],
            get_field_source=self.field_source_resolver
        )
        self.output_anchor.init(self.record_info)
        self.record_creator = self.record_info.construct_record_creator()
        self.is_initialized = True

    def read(self, n_record_limit=-1):
        if not self.is_initialized:
            raise RuntimeError(self.alteryx_engine.xmsg("Not initialized") if self.alteryx_engine else "Not initialized")

        record_count = 0

        for record_index, record in enumerate(self.reader):
            record_info_out = self.record_info.clone()
            self.record_creator.reset()

            # add all fields
            for field_name in record:
                idx = record_info_out.get_field_num(field_name)
                field = record_info_out[idx]  # type Sdk.Field
                field_value = record[field_name]
                field_schema_type = [f for f in self.reader.schema['fields'] if f['name'] == field_name][0]['type']

                serialize_as_json = False

                if isinstance(field_schema_type, dict):
                    if field_schema_type['type'] in ['record', 'array', 'map']:
                        serialize_as_json = True
                    elif field_schema_type['type'] == 'enum':
                        # An enum is encoded by a int, representing the zero-based position of the symbol in the schema.
                        field_value = field_schema_type['symbols'][field_value]
                elif isinstance(field_schema_type, list):
                    if 'null' in field_schema_type:
                        non_null_branch = [f for f in field_schema_type if f != 'null'][0]
                        field_value = field_value[non_null_branch] if field_value[non_null_branch] is not None else None

                if serialize_as_json:
                    field_value = json.dumps(field_value)

                alteryx_helpers.set_value_on_field(field, self.record_creator, field_value, self.alteryx_engine)

            # finalize record
            finalized_record_ref = self.record_creator.finalize_record()

            # push record
            self.output_anchor.push_record(finalized_record_ref)

            # report progress
            self.report_input_stream_progress()

            # check record limit
            record_count = record_count + 1
            if 0 < n_record_limit <= record_count:
                break

        return record_count
