import AlteryxPythonSDK as Sdk
import re

import odata.property
from connectors_core.alteryx.helpers import build_record_info
from odata.entity import EntityBase

from typing import Type, Union, Iterable, Sized


class EntitySchemaReader:

    nav_prop_id_pattern = re.compile(r'_[a-zA-Z0-9_]+_value')

    def __init__(self, alteryx_engine: Sdk.AlteryxEngine, entity_cls: Type[EntityBase], max_field_size=255):
        self.max_field_size = max_field_size
        self.entity_cls = entity_cls
        self.alteryx_engine = alteryx_engine

    def get_record_info(self) -> Sdk.RecordInfo:
        return build_record_info(
            self.alteryx_engine,
            [
                {
                    'name': p['name'],
                    'description': p['display_name'],
                    # because python3 ;)
                    **self.get_sdk_type_for_property_type(p['prop'], self.max_field_size)
                } for p in self._get_entity_props()
            ]
        )

    def _get_entity_props(self):
        entity = self.entity_cls()
        properties = [
            {
                'display_name': name[1:-6] if self.nav_prop_id_pattern.fullmatch(name) else name,
                'name': name,
                'prop': prop
            } for name, prop in entity.__odata__.properties
        ]
        return properties

    def get_entity_fields(self):
        return self._get_entity_props()

    def get_primary_key_fields(self):
        entity = self.entity_cls()
        primary_keys = {name: prop for name, prop in entity.__odata__.properties if prop.primary_key}
        return primary_keys.values()

    @classmethod
    def get_sdk_type_for_property_type(cls, entity_property: odata.property.PropertyBase, max_field_size=255) -> dict:
        try:
            return {
                odata.property.BooleanProperty: {'type': Sdk.FieldType.bool, 'size': 0, 'scale': 0},
                odata.property.IntegerProperty: {'type': Sdk.FieldType.int64, 'size': 0, 'scale': 0},
                odata.property.FloatProperty: {'type': Sdk.FieldType.float, 'size': 0, 'scale': 0},
                odata.property.DecimalProperty: {'type': Sdk.FieldType.fixeddecimal, 'size': 100, 'scale': 2},
                odata.property.StringProperty: {'type': Sdk.FieldType.v_wstring, 'size': max_field_size, 'scale': 0},
                odata.property.DatetimeProperty: {'type': Sdk.FieldType.datetime, 'size': 0, 'scale': 0},
                # UUID = 32 hex characters with 4 dashes
                odata.property.UUIDProperty: {'type': Sdk.FieldType.string, 'size': 36, 'scale': 0},
                odata.property.BinaryProperty: {'type': Sdk.FieldType.blob, 'size': max_field_size, 'scale': 0},
            }[type(entity_property)]
        except KeyError:
            return {'type': Sdk.FieldType.wstring, 'size': max_field_size, 'scale': 0}
