import io
from unittest.mock import MagicMock


def iterable_to_stream(iterable: iter, buffer_size: int = io.DEFAULT_BUFFER_SIZE):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.

    Lifted from https://stackoverflow.com/a/20260030
    """

    class IterStream(io.RawIOBase):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.leftover = None

        def readable(self):
            return True

        def readinto(self, b):
            try:
                length = len(b)  # We're supposed to return at most this much
                chunk = self.leftover or next(iterable)
                output, self.leftover = chunk[:length], chunk[length:]
                b[:len(output)] = output
                return len(output)
            except StopIteration:
                return 0  # indicate EOF

        def tell(self, *args, **kwargs):
            return 0

    return io.BufferedReader(IterStream(), buffer_size=buffer_size)


def get_mock_alteryx_engine(update_only_mode=False):
    """
    Get a MagicMock for the Alteryx Engine interface that is setup to provide a mocked response to get_init_var

    :param update_only_mode: the value to be returned as the mock response to engine.get_init_var('<any_tool_id>', 'UpdateOnly')
    :return: Mock for AlteryxEngineInterface
    """
    mock_engine = MagicMock()
    # mock a response to the get_init_var method, regardless of tool_id
    mock_engine.return_value.get_init_var.side_effect = (
        lambda tool_id, k: str(update_only_mode) if k == 'UpdateOnly' else ''
    )
    return mock_engine
