import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx

from typing import Union
import json

from .connections import BigQueryIncomingConnectionsHandler
from google.oauth2 import service_account, credentials
from .common import BIG_QUERY_SCOPE, OAUTH_AUD, BIG_QUERY_API_BASE


class BigQueryOutputEngine(alteryx.AyxPluginBase):

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.auth_type = None   # type: str
        self.service_account_json = None # type: str
        self.project_id = None # type: str
        self.dataset_id = None # type: str
        self.table_id = None # type: str
        self.credentials = None # type: google.auth.credentials.Credentials
        self.row_insert_batch_size = None
        self.max_repeated_records = None
        self.max_field_length = None
        self.incoming_connection_added = False

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        if self.incoming_connection_added:
            self.display_error_msg(self.alteryx_engine.xmsg(
                "Only one incoming connection can be accepted by this tool"
            ))
            return None
        connection_handler = BigQueryIncomingConnectionsHandler(self)
        self.incoming_connection_added = True
        return connection_handler

    def init_end_user_credentials(self, str_xml: str):
        access_token_str = self.get_encrypted_xml_config_value(str_xml, 'accessToken')
        try:
            access_token = json.loads(access_token_str) if access_token_str else None
        except json.decoder.JSONDecodeError as e:
            self.invalidate_config(self.alteryx_engine.xmsg("Access token is not valid JSON"))
            return False
        if not access_token:
            self.invalidate_config(self.alteryx_engine.xmsg("No access token provided for end user authentication"))
            return False

        client_id = self.get_xml_config_value(str_xml, 'clientId')
        client_secret = self.get_encrypted_xml_config_value(str_xml, 'clientSecret')
        if not client_id:
            self.invalidate_config(self.alteryx_engine.xmsg('Please specify a Client ID for end user authentication'))
            return False
        if not client_secret:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Please specify a Client secret for end user authentication'
            ))
            return False

        try:
            self.credentials = credentials.Credentials(
                access_token['access_token'],
                refresh_token=access_token['refresh_token'],
                token_uri=OAUTH_AUD,
                client_id=client_id,
                client_secret=client_secret
            )
        except ValueError as e:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Invalid access token or client credentials given for end user authentication'
            ))
            return False
        except Exception as e:
            self.invalidate_config(self.alteryx_engine.xmsg('Error authenticating to Google BigQuery'))
            return False
        return True

    def init_service_principal_credentials(self, str_xml: str):
        json_str = self.get_encrypted_xml_config_value(str_xml, 'serviceCredsJSONStr')
        try:
            self.service_account_json = json.loads(json_str) if json_str else None
        except json.decoder.JSONDecodeError as e:
            self.invalidate_config("Invalid service account keyfile.")
            return False
        if not self.service_account_json:
            self.invalidate_config(self.alteryx_engine.xmsg(
                    'Service account JSON not provided. Please specify a path to your service account file.'
            ))
            return False
        try:
            self.credentials = service_account.Credentials.from_service_account_info(
                self.service_account_json,
                scopes=[BIG_QUERY_SCOPE]
            )
        except ValueError as e:
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid service account keyfile. @1", str(e)))
            return False
        return True

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)
        self.auth_type = self.get_xml_config_value(str_xml, 'authType')
        if self.auth_type == 'EndUser':
            if not self.init_end_user_credentials(str_xml):
                return
        elif self.auth_type == 'ServicePrincipal':
            if not self.init_service_principal_credentials(str_xml):
                return
        else:
            self.invalidate_config(self.alteryx_engine.xmsg('Invalid authentication mode selected: @1', self.auth_type))
            return

        selected_id = self.get_xml_config_value(str_xml, 'selectedId')
        try:
            self.project_id, remainder = selected_id.split(':')
            self.dataset_id, self.table_id = remainder.split('.')
        except (AttributeError, ValueError) as e:
            self.invalidate_config(self.alteryx_engine.xmsg('Invalid table selection: please select a table.'))
        try:
            self.row_insert_batch_size = int(self.get_xml_config_value(str_xml, 'rowInsertBatchSize'))
        except TypeError:
            self.row_insert_batch_size = 500

        metadata_dict = {
            'Host': BIG_QUERY_API_BASE,
            'ServiceName': 'BigQuery',
            'PluginName': 'BigQuery',
            'Database': self.dataset_id,
            'Table': self.table_id
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)
        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        self.display_error_msg(self.alteryx_engine.xmsg('This tool does not allow outgoing connections'))
        return False

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        return False

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        if self.output_anchor:
            self.output_anchor.assert_close()
