import AlteryxPythonSDK as Sdk
import datetime
import decimal

from typing import Any, Callable, Sequence, Union

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
TIME_FORMAT = '%H:%M:%S'
DATE_FORMAT = '%Y-%m-%d'


def get_value_for_field(
        field: Sdk.Field,
        record_ref: Sdk.RecordRef,
        data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None) -> Any:

    def get_time_value(ref: Sdk.RecordRef) -> datetime.timedelta:
        date = datetime.datetime.strptime(field.get_as_string(ref), TIME_FORMAT)
        return datetime.timedelta(hours=date.hour, minutes=date.minute, seconds=date.second)

    accessor = {
        Sdk.FieldType.string: field.get_as_string,
        Sdk.FieldType.wstring: field.get_as_string,
        Sdk.FieldType.v_wstring: field.get_as_string,
        Sdk.FieldType.v_string: field.get_as_string,
        Sdk.FieldType.int32: field.get_as_int32,
        Sdk.FieldType.int64: field.get_as_int64,
        Sdk.FieldType.double: field.get_as_double,
        Sdk.FieldType.float: field.get_as_double,
        Sdk.FieldType.bool: field.get_as_bool,
        Sdk.FieldType.datetime: lambda ref: datetime.datetime.strptime(field.get_as_string(ref), DATETIME_FORMAT),
        Sdk.FieldType.date: lambda ref: datetime.datetime.strptime(field.get_as_string(ref), DATE_FORMAT),
        Sdk.FieldType.time: get_time_value,
        Sdk.FieldType.fixeddecimal: lambda ref: decimal.Decimal(field.get_as_string(ref)),
        Sdk.FieldType.blob: field.get_as_blob
    }[field.type]

    try:
        return accessor(record_ref)
    except TypeError as e:
        if data_conversion_error_handler:
            return data_conversion_error_handler(field, record_ref, e)
        else:
            raise


def set_value_on_field(field: Sdk.Field, record_creator: Sdk.RecordCreator, value: any,
                       alteryx_engine: Sdk.AlteryxEngine = None):
    if value is None:
        field.set_null(record_creator)
    else:
        def set_from_datetime(creator: Sdk.RecordCreator, val: Union[datetime.datetime, datetime.date, datetime.timedelta]):
            str_val = None
            if isinstance(val, datetime.datetime):
                str_val = val.strftime(DATETIME_FORMAT)
            elif isinstance(val, datetime.date):
                str_val = val.strftime(DATE_FORMAT)
            elif isinstance(val, datetime.timedelta):
                str_val = str(val)  # TODO: Should probably format like so: https://stackoverflow.com/a/30536361
            else:
                str_val = str(val)
            field.set_from_string(creator, str_val)

        def set_from_decimal(creator: Sdk.RecordCreator, val: decimal.Decimal):
            field.set_from_double(creator, float(val))

        mutate = {
            Sdk.FieldType.string: field.set_from_string,
            Sdk.FieldType.v_string: field.set_from_string,
            Sdk.FieldType.v_wstring: field.set_from_string,
            Sdk.FieldType.wstring: field.set_from_string,
            Sdk.FieldType.bool: field.set_from_bool,
            Sdk.FieldType.byte: field.set_from_int32,
            Sdk.FieldType.int16: field.set_from_int32,
            Sdk.FieldType.int32: field.set_from_int32,
            Sdk.FieldType.int64: field.set_from_int64,
            Sdk.FieldType.float: field.set_from_double,
            Sdk.FieldType.double: field.set_from_double,
            Sdk.FieldType.fixeddecimal: set_from_decimal,
            Sdk.FieldType.date: set_from_datetime,
            Sdk.FieldType.datetime: set_from_datetime,
            Sdk.FieldType.time: set_from_datetime,
            Sdk.FieldType.blob: field.set_from_blob
        }[field.type]
        try:
            mutate(record_creator, value)
        except Exception as e:
            if not alteryx_engine:
                raise ValueError(
                    "Unable to set value on field {} of type {} using value of type {}, and value = {}".format(
                        field.name, field.type, type(value), value),
                        e
                )
            else:
                raise ValueError(alteryx_engine.xmsg(
                    "Unable to set value on field @1 of type @2 using value of type @3, and value = @4",
                        field.name, field.type, type(value), str(value)
                ))



def build_record_info(alteryx_engine: Sdk.AlteryxEngine,
                      record_fields: Sequence[dict],
                      get_field_source: Callable[[Sdk.Field], str] = lambda field: '') -> Sdk.RecordInfo:
    """
    A non-interface helper for pi_push_all_records() responsible for creating the outgoing record layout.
    :param alteryx_engine:
    :param get_field_source: A function to call that will return a string containing the field source information
    :param record_fields: The name for csv file reader.
    :return: The outgoing record layout, otherwise nothing.
    """

    record_info_out = Sdk.RecordInfo(alteryx_engine)  # A fresh record info object for outgoing records.
    try:  # Add metadata info that is passed to tools downstream.
        for field in record_fields:
            record_info_out.add_field(field_name=field['name'],
                                      field_type=field['type'],
                                      size=field['size'],
                                      scale=field['scale'] or 0,
                                      source=get_field_source(field),
                                      description=field['description'] or '')
    except UnicodeError:
        raise RuntimeError(alteryx_engine.xmsg('Field names must be utf-8'))
    return record_info_out
