from abc import abstractmethod
from enum import Enum

# Use this for Azure AD authentication
from azure.datalake.store.lib import DataLakeCredential
from msrestazure.azure_active_directory import AADTokenCredentials

# Required for Azure Data Lake Store account management
from azure.mgmt.datalake.store import DataLakeStoreAccountManagementClient
from azure.mgmt.datalake.store.models import DataLakeStoreAccount

# Required for Azure Data Lake Store filesystem management
from azure.datalake.store import core, lib, multithread

# Common Azure imports
import adal
from azure.mgmt.resource.resources import ResourceManagementClient
from azure.mgmt.resource.resources.models import ResourceGroup

# Use these as needed for your application
import logging, getpass, pprint, uuid, time

AZURE_AUTHORITY_URI = 'https://login.microsoftonline.com'
DATA_LAKE_TOKEN_AUDIENCE = 'https://datalake.azure.net/'
RESOURCE_MANAGER_TOKEN_AUDIENCE = 'https://management.core.windows.net/'


class AzureAuthenticationType(Enum):
    Unknown = 0,
    ServicePrincipal = 1,
    EndUser = 2


class AdlAuthenticationBase:

    def __init__(self, tenant_id: str):
        self.tenant_id = tenant_id

    @abstractmethod
    def get_credentials(self) -> DataLakeCredential:
        pass


class AdlSpAuthentication(AdlAuthenticationBase):

    def __init__(self, store_name: str, tenant_id: str, client_id: str, secret_key: str):
        super().__init__(tenant_id)
        self.secret_key = secret_key
        self.client_id = client_id
        self.store_name = store_name

    def get_credentials(self) -> DataLakeCredential:
        return lib.auth(
            tenant_id=self.tenant_id,
            client_secret=self.secret_key,
            client_id=self.client_id,
            resource=DATA_LAKE_TOKEN_AUDIENCE)


class AdlEndUserAuthentication(AdlAuthenticationBase):

    def __init__(self, tenant_id: str, username: str, password: str):
        super().__init__(tenant_id)
        self.username = username
        self.password = password

    def get_credentials(self) -> DataLakeCredential:
        lib.Authentication()
        return lib.auth(
            username=self.username,
            password=self.password,
            tenant_id=self.tenant_id,
            resource=DATA_LAKE_TOKEN_AUDIENCE
        )

