import csv
import json
import io

import fastavro.write as avro

import azure.datalake.store as adl
import AlteryxPythonSDK as Sdk

from azure.datalake.store.exceptions import DatalakeRESTException

import connectors_core.file.errors as fileerrors
import connectors_core.file.interfaces as interfaces
import connectors_core.file.avro as avro_helper
import connectors_core.alteryx.helpers as alteryx_helpers

from abc import abstractmethod
from typing import IO, Callable, Any

from connectors_core.file import FileExistsBehavior


class ADLFileWriterBase(interfaces.FileWriter):

    def __init__(
            self,
            adl_fs: adl.AzureDLFileSystem,
            file_path: str,
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__()
        self.file_exists_behavior = file_exists_behavior
        self.data_conversion_handler = data_conversion_error_handler
        self.adl_fs = adl_fs
        self.file_path = file_path
        self.stream = None  # type: IO
        self._is_open = False
        self.records_written = 0
        self.record_info = None  # type: Sdk.RecordInfo
        self.alteryx_engine = alteryx_engine

    def init(self, record_info: Sdk.RecordInfo):
        self.record_info = record_info

    def open(self):
        try:
            file_exists = self.file_exists()
            if file_exists and self.file_exists_behavior == FileExistsBehavior.Abort:
                if self.alteryx_engine:
                    raise fileerrors.FileWriterAbortedError(self.alteryx_engine.xmsg(
                        'A file exists at the destination file path '
                        'and Existing File Handling for this output is configured to Abort'
                    ))
                else:
                    raise fileerrors.FileWriterAbortedError(
                        'A file exists at the destination file path '
                        'and Existing File Handling for this output is configured to Abort'
                    )

            file_mode = 'wb' if not file_exists or self.file_exists_behavior == FileExistsBehavior.Overwrite else 'ab'
            self.stream = self.adl_fs.open(self.file_path, mode=file_mode)
            self._is_open = True
        except DatalakeRESTException as e:
            raise fileerrors.FileOpenError(e)

    def write(self, record: Sdk.RecordRef):
        if not self.stream:
            if self.alteryx_engine:
                raise fileerrors.FileOpenError(self.alteryx_engine.xmsg(
                    "open() must be called on the writer before records are written"
                ))
            else:
                raise fileerrors.FileOpenError("open() must be called on the writer before records are written")
        self.stream.write(self.get_bytes(record))
        self.stream.flush()
        self.records_written += 1
        return True

    def flush(self):
        if self.stream:
            self.stream.flush()

    def close(self):
        self.flush()
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.adl_fs:
            self.adl_fs = None

    def is_open(self):
        return self._is_open

    def file_exists(self):
        try:
            file_info = self.adl_fs.info(path=self.file_path)
            return file_info is not None
        except FileNotFoundError:
            return False

    @abstractmethod
    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        pass


class ADLDelimitedFileWriter(ADLFileWriterBase):

    def __init__(
            self,
            adl_fs: adl.AzureDLFileSystem,
            file_path: str,
            encoding: str,
            write_header: bool = True,
            delimiter: str = ',',
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__(adl_fs, file_path, data_conversion_error_handler, file_exists_behavior, alteryx_engine)
        self.delimiter = delimiter
        self.write_header = write_header
        self.encoding = encoding

    def init(self, record_info: Sdk.RecordInfo):
        super().init(record_info)

    def write(self, record: Sdk.RecordRef):
        if self.write_header and self.records_written == 0:
            self.write_header_row()
        return super().write(record)

    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        values = [
            alteryx_helpers.get_value_for_field(field, record, self.data_conversion_handler)
            for field in self.record_info
        ]  # type: Sdk.Field
        with io.StringIO('') as temp_stream:
            self.get_csv_writer(temp_stream).writerow(values)
            return temp_stream.getvalue().encode(self.encoding)

    def write_header_row(self):
        field_names = []
        for field in self.record_info:  # type: Sdk.Field
            field_names.append(field.name)
        with io.StringIO('') as temp_stream:
            self.get_csv_writer(temp_stream).writerow(field_names)
            self.stream.write(temp_stream.getvalue().encode(self.encoding))

    def get_csv_writer(self, stream: IO) -> csv.writer:
        return csv.writer(stream, delimiter=self.delimiter)


class ADLAvroFileWriter(ADLFileWriterBase):

    def __init__(
            self,
            adl_fs: adl.AzureDLFileSystem,
            file_path: str,
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__(adl_fs, file_path, data_conversion_error_handler, file_exists_behavior, alteryx_engine)
        self.schema = None  # type: dict
        self.writer = None  # type: avro.Writer

    def init(self, record_info: Sdk.RecordInfo):
        super().init(record_info)
        self.schema = {
            'doc': 'Alteryx Azure Data Lake File Output',
            'name': 'AlteryxADLOutput',
            'namespace': 'alteryx.adl',
            'type': 'record',
            'fields': [
                {'name': f.name, 'type': avro_helper.get_avro_type_for_sdk_type(f.type, self.alteryx_engine)}
                for f in record_info
            ]
        }

    def open(self):
        super().open()
        self.writer = avro.Writer(self.stream, self.schema)

    def flush(self):
        if self.writer:
            self.writer.dump()
        super().flush()

    def close(self):
        super().close()
        self.writer = None

    def write(self, record: Sdk.RecordRef):
        self.writer.write(
            {
                f.name:
                    avro_helper.get_avro_value_from_field(f, record, self.data_conversion_handler, self.alteryx_engine)
                    for f in self.record_info
            }
        )
        return True

    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        raise NotImplementedError()  # write method is overridden


class ADLJSONFileWriter(ADLFileWriterBase):
    def __init__(
            self,
            adl_fs: adl.AzureDLFileSystem,
            file_path: str,
            data_conversion_error_handler: Callable[[Sdk.Field, Sdk.RecordRef, BaseException], Any] = None,
            file_exists_behavior: FileExistsBehavior=FileExistsBehavior.Abort,
            alteryx_engine: Sdk.AlteryxEngine = None
    ):
        super().__init__(adl_fs, file_path, data_conversion_error_handler, file_exists_behavior, alteryx_engine)

    def open(self):
        super().open()
        self.stream.write(b'[\n')  # Write start array

    def write(self, record: Sdk.RecordRef):
        if self.records_written > 0:
            self.stream.write(b',\n')
        return super().write(record)

    def get_bytes(self, record: Sdk.RecordRef) -> bytes:
        return json.dumps(
            {
                f.name:
                    alteryx_helpers.get_value_for_field(f, record, self.data_conversion_handler)
                    for f in self.record_info
            }
        ).encode()

    def close(self):
        if self.stream:
            self.stream.write(b']')  # write end array
        super().close()
